//
//  MeshBuilder.swift
//  MetalRenderer
//
//  Created by Caroline on 20/2/20.
//  Copyright © 2020 Razeware. All rights reserved.
//

import MetalKit
import Engine

enum PrimitiveType {
  case rect
}

class Primitive: Node, Renderable {
  struct Vertex {
    var position: float3
  }

  struct Mesh {
    let vertexBuffer: MTLBuffer
    let submeshes: [Submesh]
  }

  struct Submesh {
    let indexBuffer: MTLBuffer
    let indexCount: Int
    let indexType: MTLIndexType
  }
  
  let pipelineState: MTLRenderPipelineState
  let meshes: [Mesh]
  var color: float3 = [0, 0, 0]
 
  init(_ primitiveType: PrimitiveType,
       size: float3) {
    
    switch primitiveType {
    case .rect:
      let mesh = Self.buildRect(size: size)
      self.meshes = [mesh]
    }
    pipelineState = Self.createPipelineState()
    super.init()
    
    let halfSize = Self.half(size)
    let minBounds = float3(-halfSize.x, -halfSize.y, -halfSize.z)
    let maxBounds = float3(halfSize.x, halfSize.y, halfSize.z)
    self.boundingBox = MDLAxisAlignedBoundingBox(maxBounds: maxBounds,
                                                 minBounds: minBounds)
  }
  
  static func half(_ size: float3) -> float3 {
    [size.x * 0.5, size.y * 0.5, size.z * 0.5]
  }
  
  static func buildRect(size: float3) -> Mesh {
    var vertices: [Vertex] {
      let x = half(size).x
      let y = half(size).y
      let z = half(size).z
      let vertices =
     [
      Vertex(position: [-x,  y, z]),
      Vertex(position: [ x,  y, z]),
      Vertex(position: [ x, -y, z]),
      Vertex(position: [-x, -y, z])
      ]
      return vertices
    }
    
    var indices: [UInt16] = [
      0, 1, 2,
      0, 2, 3
    ]
    let vertexBuffer = Renderer.device.makeBuffer(bytes: vertices,
                                                  length: MemoryLayout<Vertex>.stride * vertices.count,
                                                  options: [])!
    let indexBuffer = Renderer.device.makeBuffer(bytes: &indices,
                                                 length: MemoryLayout<UInt16>.stride * indices.count,
                                                 options: [])!
    let submesh = Submesh(indexBuffer: indexBuffer,
                            indexCount: indices.count,
                            indexType: .uint16)
    let mesh = Mesh(vertexBuffer: vertexBuffer, submeshes: [submesh])
    return mesh
  }

  
  static func createPipelineState() -> MTLRenderPipelineState {
    let library = Renderer.device.makeDefaultLibrary()!
    let vertexFunction = library.makeFunction(name: "vertex_new")
    let fragmentFunction = library.makeFunction(name: "fragment_new")
    
    let pipelineDescriptor = MTLRenderPipelineDescriptor()
    pipelineDescriptor.colorAttachments[0].pixelFormat = .bgra8Unorm
    pipelineDescriptor.depthAttachmentPixelFormat = .depth32Float
    
    pipelineDescriptor.vertexFunction = vertexFunction
    pipelineDescriptor.fragmentFunction = fragmentFunction
    
    let pipelineState: MTLRenderPipelineState
    do {
      pipelineState = try Renderer.device.makeRenderPipelineState(descriptor: pipelineDescriptor)
    } catch {
      fatalError("Pipeline state not created: \(error.localizedDescription)")
    }
    return pipelineState
  }

  
  func render(renderEncoder: MTLRenderCommandEncoder, uniforms vertex: Uniforms, fragmentUniforms fragment: FragmentUniforms) {
    var uniforms = vertex
    uniforms.modelMatrix = worldMatrix
    var fragmentUniforms = fragment
    renderEncoder.setFragmentBytes(&fragmentUniforms,
                                   length: MemoryLayout<FragmentUniforms>.stride,
                                   index: 22)
    renderEncoder.setFragmentBytes(&color,
                                   length: MemoryLayout<float3>.stride,
                                   index: 22)
    renderEncoder.setVertexBytes(&uniforms,
                                 length: MemoryLayout<Uniforms>.stride,
                                 index: 21)
    renderEncoder.setRenderPipelineState(pipelineState)
    for mesh in meshes {
      renderEncoder.setVertexBuffer(mesh.vertexBuffer,
                                    offset: 0, index: 0)
      for submesh in mesh.submeshes {
        renderEncoder.drawIndexedPrimitives(type: .triangle,
                                            indexCount: submesh.indexCount,
                                            indexType: submesh.indexType,
                                            indexBuffer: submesh.indexBuffer,
                                            indexBufferOffset: 0)
      }
    }
  }
}
